@@ -3,6 +3,7 @@
 #define _POSIX_SOURCE
 #endif
 #include <stdio.h>
+#include <stdlib.h>
 #include <malloc.h>
 #include <string.h>
 #include <fcntl.h>
@@ -22,9 +23,10 @@
 #define PATH_MAX 1024
 #endif
 
-#define progver "%s: scan/change symbolic links - v1.2 - by Mark Lord\n\n"
+#define progver "%s: scan/change symbolic links - v1.3 - by Mark Lord\n\n"
 static char *progname;
-static int verbose = 0, fix_links = 0, recurse = 0, delete = 0, shorten = 0, testing = 0;
+static int verbose = 0, fix_links = 0, recurse = 0, delete = 0, shorten = 0,
+		testing = 0, single_fs = 1;
 
 /*
  * tidypath removes excess slashes and "." references from a path string
@@ -72,22 +74,25 @@ static int tidy_path (char *path)
 	int tidied = 0;
 	char *s, *p;
 
-	strcat(path,"/");	/* tmp trailing slash simplifies things */
+	s = path + strlen(path) - 1;
+	if (s[0] != '/') {	/* tmp trailing slash simplifies things */
+		s[1] = '/';
+		s[2] = '\0';
+	}
 	while (substr(path, "/./", "/"))
 		tidied = 1;
 	while (substr(path, "//", "/"))
 		tidied = 1;
 
-	while (NULL != (p = strstr(path,"/../"))) {
-		for (s = p+3; p != path && *--p != '/';);
+	while ((p = strstr(path,"/../")) != NULL) {
+		s = p+3;
+		for (p--; p != path; p--) if (*p == '/') break;
 		if (*p != '/')
 			break;
-		if (p == path) 
-			++p;
 		while ((*p++ = *s++));
 		tidied = 1;
 	}
-	if (!*path)
+	if (*path == '\0')
 		strcpy(path,"/");
 	p = path + strlen(path) - 1;
 	if (p != path && *p == '/')
@@ -152,7 +157,7 @@ static void fix_symlink (char *path, dev_t my_dev)
 {
 	static char lpath[PATH_MAX], new[PATH_MAX], abspath[PATH_MAX];
 	char *p, *np, *lp, *tail, *msg;
-	struct stat stbuf;
+	struct stat stbuf, lstbuf;
 	int c, fix_abs = 0, fix_messy = 0, fix_long = 0;
 
 	if ((c = readlink(path, lpath, sizeof(lpath))) == -1) {
@@ -165,10 +170,11 @@ static void fix_symlink (char *path, dev_t my_dev)
 	abspath[0] = '\0';
 	if (lpath[0] != '/') {
 		strcat(abspath,path);
-		if (((c = strlen(abspath)) > 0) && (abspath[c-1] == '/'))
-			abspath[c-1] = '\0';
+		c = strlen(abspath);
+		if ((c > 0) && (abspath[c-1] == '/'))
+			abspath[c-1] = '\0'; /* cut trailing / */
 		if ((p = strrchr(abspath,'/')) != NULL)
-			*p = '\0';
+			*p = '\0'; /* cut last component */
 		strcat(abspath,"/");
 	}
 	strcat(abspath,lpath);
@@ -185,7 +191,11 @@ static void fix_symlink (char *path, dev_t my_dev)
 		}
 		return;
 	}
-	if (stbuf.st_dev != my_dev) {
+
+	if (single_fs)
+		lstat(abspath, &lstbuf); /* if the above didn't fail, then this shouldn't */
+	
+	if (single_fs && lstbuf.st_dev != my_dev) {
 		msg = "other_fs:";
 	} else if (lpath[0] == '/') {
 		msg = "absolute:";
@@ -234,7 +244,9 @@ static void fix_symlink (char *path, dev_t my_dev)
 		}
 		strcpy (np, tail);
 		(void) tidy_path(new);
+		if (shorten) (void) shorten_path(new, path);
 	}
+	shorten_path(new,path);
 	if (!testing) {
 		if (unlink (path)) {
 			perror(path);
@@ -285,16 +297,19 @@ static void dirwalk (char *path, int pathlen, dev_t dev)
 static void usage_error (void)
 {
 	fprintf(stderr, progver, progname);
-	fprintf(stderr, "Usage:\t%s [-crsv] dirlist\n\n", progname);
-	fprintf(stderr, "Flags:\t-c == change absolute/messy links to relative\n"
+	fprintf(stderr, "Usage:\t%s [-cdorstv] dirlist\n\n", progname);
+	fprintf(stderr, "Flags:"
+		"\t-c == change absolute/messy links to relative\n"
 		"\t-d == delete dangling links\n"
+		"\t-o == warn about links across file systems\n"
 		"\t-r == recurse into subdirs\n"
-		"\t-s == shorten lengthy links\n"
+		"\t-s == shorten lengthy links (displayed in output only when -c not specified)\n"
+		"\t-t == show what would be done by -c\n"
 		"\t-v == verbose (show all symlinks)\n\n");
 	exit(1);
 }
 
-void main(int argc, char **argv)
+int main(int argc, char **argv)
 {
 	static char path[PATH_MAX+2], cwd[PATH_MAX+2];
 	int dircount = 0;
@@ -318,12 +333,13 @@ void main(int argc, char **argv)
 			if (*++p == '\0')
 				usage_error();
 			while ((c = *p++)) {
-				     if (c == 'v')	verbose   = 1;
-				else if (c == 'c')	fix_links = 1;
-				else if (c == 'r')	recurse   = 1;
+				     if (c == 'c')	fix_links = 1;
 				else if (c == 'd')	delete    = 1;
+				else if (c == 'o')	single_fs = 0;
+				else if (c == 'r')	recurse   = 1;
 				else if (c == 's')	shorten   = 1;
 				else if (c == 't')	testing   = 1;
+				else if (c == 'v')	verbose   = 1;
 				else			usage_error();
 			}
 		} else {
